/**********************************************************************
 *
 * cgiwrapper.cpp -- output pages using the cgi protocol
 * Copyright (C) 1999  The New Zealand Digital Library Project
 *
 * A component of the Greenstone digital library software
 * from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: cgiwrapper.cpp 963 2000-02-21 22:01:02Z sjboddie $
 *
 *********************************************************************/

/*
   $Log$
   Revision 1.22  2000/02/21 21:56:46  sjboddie
   gsdlhome now comes from gsdlsite.cfg

   Revision 1.21  2000/01/25 22:45:59  sjboddie
   few changes to get fastcgi to work properly

   Revision 1.20  1999/09/07 04:56:53  sjboddie
   added GPL notice

   Revision 1.19  1999/09/02 00:24:36  rjmcnab
   fixed bug in getting POST arguments

   Revision 1.18  1999/08/20 01:02:07  sjboddie
   added some usage logging

   Revision 1.17  1999/07/15 06:03:15  rjmcnab
   Moved the adding of the actions to librarymain so that they can be
   overriden easily.

   Revision 1.16  1999/07/14 08:31:05  rjmcnab
   Fixed a small bug in the POST implementation.

   Revision 1.15  1999/07/13 23:32:17  rjmcnab
   Added authenaction and usersaction

   Revision 1.14  1999/07/11 01:03:37  rjmcnab
   Added ability to receive POST cgi form data.

   Revision 1.13  1999/06/24 05:12:18  sjboddie
   lots of small changes

   Revision 1.12  1999/04/30 01:59:40  sjboddie
   lots of stuff - getting documentaction working (documentaction replaces
   old browseaction)

   Revision 1.11  1999/03/25 03:12:01  sjboddie

   subjectbrowseaction was replaced with browseaction

   Revision 1.10  1999/03/05 03:53:54  sjboddie

   fixed some bugs

   Revision 1.9  1999/03/04 22:38:21  sjboddie

   Added subjectbrowseaction. - Doesn't do anything yet.

   Revision 1.8  1999/02/28 20:00:13  rjmcnab


   Fixed a few things.

   Revision 1.7  1999/02/21 22:33:53  rjmcnab

   Lots of stuff :-)

   Revision 1.6  1999/02/12 02:40:17  sjboddie

   Added page action

   Revision 1.5  1999/02/11 01:24:04  rjmcnab

   Fixed a few compiler warnings.

   Revision 1.4  1999/02/08 01:28:01  rjmcnab

   Got the receptionist producing something using the statusaction.

   Revision 1.3  1999/02/05 10:42:44  rjmcnab

   Continued working on receptionist

   Revision 1.2  1999/02/04 10:00:56  rjmcnab

   Developed the idea of an "action" and having them define the cgi arguments
   which they need and how those cgi arguments function.

   Revision 1.1  1999/02/04 01:16:17  rjmcnab

   Initial revision.

   Revision 1.5  1999/01/19 01:38:18  rjmcnab

   Made the source more portable.

   Revision 1.4  1999/01/12 01:51:04  rjmcnab

   Standard header.

 */


#include "gsdlconf.h"
#include "cgiwrapper.h"
#include "recptconfig.h"
#include "fileutil.h"
#include <stdlib.h>
#include <assert.h>

#if defined(GSDL_USE_OBJECTSPACE)
#  include <ospace/std/iostream>
#  include <ospace/std/fstream>
#elif defined(GSDL_USE_IOS_H)
#  include <iostream.h>
#  include <fstream.h>
#else
#  include <iostream>
#  include <fstream>
#endif

#ifdef USE_FASTCGI
#include "fcgiapp.h"
#endif


#ifdef USE_FASTCGI
// used to output the text from receptionist
class fcgistreambuf : public streambuf {
public:
  fcgistreambuf ();
  int sync ();
  int overflow (int ch);
  int underflow () {return EOF;}
  
  void fcgisbreset() {fcgx_stream = NULL; other_ostream = NULL;};
  void set_fcgx_stream(FCGX_Stream *newone) {fcgx_stream=newone;};
  void set_other_ostream(ostream *newone) {other_ostream=newone;};
  
private:
  FCGX_Stream *fcgx_stream;
  ostream *other_ostream;
};

fcgistreambuf::fcgistreambuf() {
  fcgisbreset();
  if (base() == ebuf()) allocate();
  setp (base(), ebuf());
};

int fcgistreambuf::sync () {
  if ((fcgx_stream != NULL) &&
      (FCGX_PutStr (pbase(), out_waiting(), fcgx_stream) < 0)) {
    fcgx_stream = NULL;
  }

  if (other_ostream != NULL) {
    char *thepbase=pbase();
    for (int i=0;i<out_waiting();i++) (*other_ostream).put(thepbase[i]);
  }
  
  setp (pbase(), epptr());
  
  return 0;
}

int fcgistreambuf::overflow (int ch) {
  if (sync () == EOF) return EOF;
  if (ch != EOF) sputc (ch);
  return 0;
}

#endif


static void page_errorsitecfg (text_t &errorpage, int mode) {
  errorpage += "Content-type: text/html\n\n";

  errorpage += "<html>\n";
  errorpage += "<head>\n";
  errorpage += "<title>Error</title>\n";
  errorpage += "</head>\n";
  errorpage += "<body>\n";
  errorpage += "<h2>Oops!</h2>\n";
  if (mode == 0) {
    errorpage += "The gsdlsite.cfg configuration file could not be found. This file\n";
    errorpage += "should contain configuration information relating to this\n";
    errorpage += "site's setup.\n";
    errorpage += "gsdlsite.cfg should reside in the same directory as this executable\n";
  } else if (mode == 1) {
    errorpage += "The gsdlsite.cfg configuration file does not contain a valid gsdlhome\n";
    errorpage += "entry.\n";
    errorpage += "gsdlsite.cfg resides in the same directory as this executable\n";
  }
  errorpage += "</body>\n";
  errorpage += "</html>\n";
}


static void page_errormaincfg (const text_t &gsdlhome, const text_t &collection,
			       text_t &errorpage) {
  errorpage += "Content-type: text/html\n\n";

  errorpage += "<html>\n";
  errorpage += "<head>\n";
  errorpage += "<title>Error</title>\n";
  errorpage += "</head>\n";
  errorpage += "<body>\n";
  errorpage += "<h2>Oops!</h2>\n";
  if (collection.empty()) {
    text_t main_cfg_file = filename_cat (gsdlhome, "etc", "main.cfg");
    errorpage += "The main.cfg configuration file could not be found. This file\n";
    errorpage += "should contain configuration information relating to the\n";
    errorpage += "setup of the interface. As this cgi script is not being run\n";
    errorpage += "in collection specific mode the file should reside at\n";
    errorpage += main_cfg_file + ".\n";
  } else {
    text_t collect_cfg_file = filename_cat (gsdlhome, "collect", collection, "etc", "collect.cfg");
    text_t main_collect_cfg_file = filename_cat (gsdlhome, "etc", "collect.cfg");
    text_t main_cfg_file = filename_cat (gsdlhome, "etc", "main.cfg");
    errorpage += "Either the collect.cfg or main.cfg configuration file could\n";
    errorpage += "not be found. This file should contain configuration information\n";
    errorpage += "relating to the setup of the interface. As this cgi script is\n";
    errorpage += "being run in collection specific mode the file should reside\n";
    errorpage += "at either ";
    errorpage += collect_cfg_file + ",\n";
    errorpage += main_collect_cfg_file + " or " + main_cfg_file + ".\n";
  }
  errorpage += "</body>\n";
  errorpage += "</html>\n";
}


static void page_errorinit (const text_t &gsdlhome, text_t &errorpage) {
  errorpage += "Content-type: text/html\n\n";

  errorpage += "<html>\n";
  errorpage += "<head>\n";
  errorpage += "<title>Error</title>\n";
  errorpage += "</head>\n";
  errorpage += "<body>\n";
  errorpage += "<h2>Oops!</h2>\n";
  errorpage += "An error occurred during the initialisation of the Greenstone Digital\n";
  errorpage += "Library software. It is likely that the software has not been setup\n";
  errorpage += "correctly.\n";

  text_t init_file = filename_cat (gsdlhome, "etc", "initout.txt");
  char *ifile = init_file.getcstr();
  ifstream initin (ifile);
  delete ifile;
  if (initin) {
    errorpage += "The initialisation error log, " + init_file + ", contains the\n";
    errorpage += "following information:\n\n";
    errorpage += "<pre>\n";

    char c;
    initin.get(c);
    while (!initin.eof ()) {
      errorpage.push_back(c);
      initin.get(c);
    }
    
    errorpage += "</pre>\n";

    initin.close();

  } else {
    errorpage += "Please consult " + init_file + " for more information.\n";
  }

  errorpage += "</body>\n";
  errorpage += "</html>\n";
}

static void page_errorparseargs (const text_t &gsdlhome, text_t &errorpage) {
  errorpage += "Content-type: text/html\n\n";

  errorpage += "<html>\n";
  errorpage += "<head>\n";
  errorpage += "<title>Error</title>\n";
  errorpage += "</head>\n";
  errorpage += "<body>\n";
  errorpage += "<h2>Oops!</h2>\n";
  errorpage += "An error occurred during the parsing of the cgi arguments.\n";

  text_t error_file = filename_cat (gsdlhome, "etc", "errout.txt");
  char *efile = error_file.getcstr();
  ifstream errin (efile);
  delete efile;
  if (errin) {
    errorpage += "The error log, " + error_file + ", contains the\n";
    errorpage += "following information:\n\n";
    errorpage += "<pre>\n";

    char c;
    errin.get(c);
    while (!errin.eof ()) {
      errorpage.push_back(c);
      errin.get(c);
    }
    errorpage += "</pre>\n";
    errin.close();

  } else {
    errorpage += "Please consult " + error_file + " for more information.\n";
  }

  errorpage += "</body>\n";
  errorpage += "</html>\n";
}

static void page_errorcgipage (const text_t &gsdlhome, text_t &errorpage) {
  errorpage += "Content-type: text/html\n\n";

  errorpage += "<html>\n";
  errorpage += "<head>\n";
  errorpage += "<title>Error</title>\n";
  errorpage += "</head>\n";
  errorpage += "<body>\n";
  errorpage += "<h2>Oops!</h2>\n";
  errorpage += "An error occurred during the construction of the cgi page.\n";


  text_t error_file = filename_cat (gsdlhome, "etc", "errout.txt");
  char *efile = error_file.getcstr();
  ifstream errin (efile);
  delete efile;
  if (errin) {
    errorpage += "The error log, " + error_file + ", contains the\n";
    errorpage += "following information:\n\n";
    errorpage += "<pre>\n";

    char c;
    errin.get(c);
    while (!errin.eof ()) {
      errorpage.push_back(c);
      errin.get(c);
    }
    errorpage += "</pre>\n";
    errin.close();

  } else {
    errorpage += "Please consult " + error_file + " for more information.\n";
  }

  errorpage += "</body>\n";
  errorpage += "</html>\n";
}


// cgiwrapper does everything necessary to output a page
// using the cgi protocol. If this is being run for a particular
// collection then "collection" should be set, otherwise it
// should equal "".
void cgiwrapper (receptionist &recpt, text_t collection) {
#ifdef USE_FASTCGI
  fcgistreambuf outbuf;
#endif

  // init stuff - we can't output error pages directly with
  // fastcgi so the pages are stored until we can output them
  text_t errorpage;
  outconvertclass text_t2ascii;

  // set defaults
  int maxrequests = 10000;
  recpt.configure ("collection", collection);
  recpt.configure ("httpimg", "/gsdl/images");
  char *script_name = getenv("SCRIPT_NAME");
  if (script_name != NULL) recpt.configure("gwcgi", script_name);
  else recpt.configure("gwcgi", "/cgi-bin/gw");

  // read in the configuration files.
  text_t gsdlhome;
  if (!site_cfg_read (recpt, gsdlhome, maxrequests)) {
    // couldn't find the site configuration file
    page_errorsitecfg (errorpage, 0);
  } else if (gsdlhome.empty()) {
    // no gsdlhome in gsdlsite.cfg
    page_errorsitecfg (errorpage, 1);
  } else if (!main_cfg_read (recpt, gsdlhome, collection)) {
    // couldn't find the main configuration file
    page_errormaincfg (gsdlhome, collection, errorpage);
  }

  // initialise the library software
  if (errorpage.empty()) {
    text_t init_file = filename_cat (gsdlhome, "etc", "initout.txt");
    char *iout = init_file.getcstr();
    ofstream initout (iout);
    delete iout;
    if (!recpt.init(initout)) {
      // an error occurred during the initialisation
      initout.close();
      page_errorinit(gsdlhome, errorpage);
    }
    initout.close();
  }
  
  // find out whether this is being run as a cgi-script
  // or a fastcgi script
  int numrequests = 0;
#ifdef USE_FASTCGI
  int isfastcgi = !FCGX_IsCGI();
  FCGX_Stream *fcgiin, *fcgiout, *fcgierr;
  FCGX_ParamArray fcgienvp;
#else
  int isfastcgi = 0;
#endif

  // get the query string if it is not being run as a fastcgi
  // script
  text_t argstr = "";
  cgiargsclass args;
  char *aURIStr;
  if (!isfastcgi) {
    char *request_method_str = getenv("REQUEST_METHOD");
    char *content_length_str = getenv("CONTENT_LENGTH");
    if (request_method_str != NULL && strcmp(request_method_str, "POST") == 0 &&
	content_length_str != NULL)  {
      // POST form data
      int content_length = text_t(content_length_str).getint();
      if (content_length > 0) {
	char c;
	do {
	  cin.get(c);
	  if (cin.eof()) break;
	  argstr.push_back (c);
	  content_length--;
	} while (content_length > 0);
      }
      
    } else {
      aURIStr = getenv("QUERY_STRING");
      if ((request_method_str != NULL && strcmp(request_method_str, "GET") == 0)
	  || aURIStr != NULL) {
	// GET form data
	if (aURIStr != NULL) argstr = aURIStr;
      } else {
	// debugging from command line
	char cinURIStr[1024];
	cin.get(cinURIStr, 1024);
	argstr = cinURIStr;
      }
    }

    // cgi scripts only deal with one request
    maxrequests = 1;
  }

  // Page-request loop. If this is not being run as a fastcgi
  // process then only one request will be processed and then
  // the process will exit.
  while (numrequests < maxrequests) {
#ifdef USE_FASTCGI
    if (isfastcgi) {
      if (FCGX_Accept(&fcgiin, &fcgiout, &fcgierr, &fcgienvp) < 0) break;
      aURIStr = FCGX_GetParam("QUERY_STRING", fcgienvp);
      if (aURIStr != NULL) argstr = aURIStr;
      else argstr = "";
    }
#endif

    // get output streams ready
#ifdef USE_FASTCGI
    outbuf.fcgisbreset ();
    if (isfastcgi) outbuf.set_fcgx_stream (fcgiout);
    else outbuf.set_other_ostream (&cout);
    ostream pageout (&outbuf);
#else
#define pageout cout
#endif

    // if using fastcgi we'll load environment into a map,
    // otherwise simply pass empty map (can't get environment
    // variables using getenv() while using FCGX versions
    // of fastcgi - at least I can't ;-) - Stefan)
    text_tmap fastcgienv;
#ifdef USE_FASTCGI
    if (isfastcgi) {
      for(; *fcgienvp != NULL; fcgienvp++) {
	text_t fvalue = *fcgienvp;
	text_t::const_iterator begin = fvalue.begin();
	text_t::const_iterator end = fvalue.end();
	text_t::const_iterator equals_sign = findchar (begin, end, '=');
	if (equals_sign != end)
	  fastcgienv[substr(begin, equals_sign)] = substr(equals_sign+1, end);
      }
    }
#endif

    // temporarily need to configure gwcgi here when using fastcgi as I can't
    // get it to pass the SCRIPT_NAME environment variable to the initial 
    // environment (if anyone can work out how to do this using the apache 
    // server, let me know). Note that this overrides the gwcgi field in 
    // site.cfg (which it shouldn't do) but I can't at present set gwcgi 
    // from site.cfg as I have old receptionists laying around that wouldn't
    // appreciate it. The following 5 lines of code should be deleted once
    // I either a: get the server to pass SCRIPT_NAME at initialization
    // time or b: convert all the collections using old receptionists over
    // to this version and uncomment gwcgi in the site.cfg file -- Stefan.
#ifdef USE_FASTCGI
    if (isfastcgi) {
      recpt.configure("gwcgi", fastcgienv["SCRIPT_NAME"]);
    }
#endif


    if (errorpage.empty()) {
      text_t error_file = filename_cat (gsdlhome, "etc", "errout.txt");
      char *eout = error_file.getcstr();
      ofstream errout (eout);
      delete eout;
      cerr = errout;
    
      // parse the cgi arguments and produce the resulting page if there
      // has been no errors so far
      if (!recpt.parse_cgi_args (argstr, args, errout, fastcgienv)) {
	errout.close ();
	page_errorparseargs(gsdlhome, errorpage);
      } else {
	if (!recpt.produce_cgi_page (args, pageout, errout, fastcgienv)) {
	  errout.close ();
	  page_errorcgipage(gsdlhome, errorpage);
	} else {
	  errout.close ();
	}
	recpt.log_cgi_args (args, errout, fastcgienv);
      }
    }
    // there was an error, output the error page
    if (!errorpage.empty()) {
      pageout << text_t2ascii << errorpage;
      errorpage.clear();
      numrequests = maxrequests; // make this the last page
    }
    pageout << flush;
    
    // finish with the output streams
#ifdef USE_FASTCGI
    if (isfastcgi) FCGX_Finish();
#endif

    numrequests++;
  }

  return;
}

